from cryptography.fernet import Fernet

def generate_key():
    key = Fernet.generate_key()
    file = open("/Users/liuzijun1/Desktop/Librec-auto/librec-auto-library/librec_auto/librec_auto/cmd/post/key.key", "wb")
    file.write(key)
    file.close()

def encrypted_file():
    file = open("/Users/liuzijun1/Desktop/Librec-auto/librec-auto-library/librec_auto/librec_auto/cmd/post/key.key", "rb")
    key = file.read()
    file.close()

    with open("/Users/liuzijun1/Desktop/Librec-auto/librec-auto-library/librec_auto/librec_auto/cmd/post/Key_API.txt", "rb") as api:
        data = api.read()
    fernet = Fernet(key)
    encrypted = fernet.encrypt(data)
    with open("/Users/liuzijun1/Desktop/Librec-auto/librec-auto-library/librec_auto/librec_auto/cmd/post/Key_API.txt.encrypted", "wb") as encry:
        encry.write(encrypted)

def encrypted_file_dropbox():
    file = open("/Users/liuzijun1/Desktop/Librec-auto/librec-auto-library/librec_auto/librec_auto/cmd/post/key.key", "rb")
    key = file.read()
    file.close()

    with open("/Users/liuzijun1/Desktop/Librec-auto/librec-auto-library/librec_auto/librec_auto/cmd/post/Key_API_Dropbox.txt", "rb") as api:
        data = api.read()
    fernet = Fernet(key)
    encrypted = fernet.encrypt(data)
    with open("/Users/liuzijun1/Desktop/Librec-auto/librec-auto-library/librec_auto/librec_auto/cmd/post/Key_API_Dropbox.txt.encrypted", "wb") as encry:
        encry.write(encrypted)

generate_key()
encrypted_file()
encrypted_file_dropbox()



from librec_auto.core.cmd import Cmd
from librec_auto.core.util import Files, SubPaths
from librec_auto.core import ConfigCmd
from  librec_auto.core.util import confirm
import shutil
import os


class PurgeCmd (Cmd):

    def __str__(self):
        return f'PurgeCmd({self._type})'

    def __init__(self, type, noask=False):
        self._type = type
        self._noask = noask
        self._files = None

    def setup(self, args):
        pass

    def dry_run(self, config):
        print(f"librec-auto (DR): Executing purge command {self}")

    def execute(self, config: ConfigCmd):
        self.status = Cmd.STATUS_INPROC
        self._files = config.get_files()

        if self._noask or self.purge_confirm():
            if self._type == "all" or self._type == 'split':
                self.purge_subexperiments()
                self.purge_splits()
                self.purge_post()

            if self._type == "results":
                self.purge_subexperiments()
                self.purge_post()

            if self._type == "rerank":
                self.purge_rerank()
                self.purge_post()

            if self._type == "post":
                self.purge_post()
        else:
            print("librec-auto: Skipping. No files deleted.")
        self.status = Cmd.STATUS_COMPLETE

    def purge_confirm(self):
        target = self._files.get_exp_path()
        prompt_str = f"This will perform a purge of type {self._type} experiments and/or file splits in directory {target}"
        return confirm(prompt=prompt_str, resp=False)


    def purge_subexperiments(self):
        target = self._files.get_exp_path()

        print ("librec-auto: Purging sub-experiments ", target)
        if self._files.get_sub_count() > 0:
            for sub_paths in self._files.get_sub_paths_iterator():
                exp_str = sub_paths.get_path_str('subexp')
                print ("librec-auto: Deleting experiment directory: ", exp_str)
                shutil.rmtree(exp_str)
        else:
            print ("librec-auto: No experiments folders found in ", target)

    def purge_splits(self):
        target = self._files.get_exp_path()
        split_path = self._files.get_split_path()
        if split_path.exists():
            print ("librec-auto: Deleting split directories ", target)
            shutil.rmtree(split_path.as_posix())
        else:
            print ("librec-auto: No split directories found in", target)

    def purge_post(self):
        target = self._files.get_exp_path()
        post_path = self._files.get_post_path()

        if post_path.exists():
            print ("librec-auto: Deleting post directory files ", target)

            files = post_path.glob('*')
            for f in files:
                os.remove(str(f))
        else:
            print ("librec-auto: Post directory missing ", target)

    # TODO: 2019-12-06 RB The status file will be out of date.
    def purge_rerank(self):
        if self._files.get_sub_count() > 0:
            for sub_paths in self._files.get_sub_paths_iterator():
                exp_str = sub_paths.get_path_str('subexp')
                print ("librec-auto: Deleting reranked results: ", exp_str)
                sub_paths.results2original()







from librec_auto.core.cmd import Cmd
from librec_auto.core.util import Files
from librec_auto.core import ConfigCmd
from librec_auto.core.util import LogFile, Status
import sys
import subprocess


class StatusCmd(Cmd):

    def __str__(self):
        return f"StatusCmd()"

    def setup(self, args):
        pass

    def dry_run(self, config):
        print (f'librec-auto (DR): Running status command {self}')

    def execute(self, config: ConfigCmd):
        self.status = Cmd.STATUS_INPROC
        files = config.get_files()
        target = files.get_exp_path()
#        result_path = files.get_result_path()

        if files.get_sub_count()==0:
            print("librec-auto: No experiments found.")
        else:
            for sub_paths in files.get_sub_paths_iterator():
                status = Status(sub_paths)
                print(status)

        self.status = Cmd.STATUS_COMPLETE





import logging
from librec_auto.core.util.xml_utils import xml_load_from_path

class LibraryColl:

    _libraries = []

    # Add to front so the last library listed is consulted first
    def add_lib (self, lib):
        self._libraries.insert(0, lib)

    def get_elem(self, name):
        for lib in self._libraries:
            print(f'Checking library {lib.lib_path}')
            elem = lib.get_elem(name)
            if elem is not None:
                return elem
        return None

class Library:

    _xml = None
    _elem_dict = {}

    lib_path = None

    # Read in the library: if src="system", then look in the rules folder
    # Otherwise look in the config folder
    def __init__(self, filename, source, files):
        if source and source == 'system':
            path = files.get_lib_path()
        elif source: # But not 'system'
            logging.warning(f'Only system src attribute supported. Got {source}')
            path = files.get_config_dir_path()
        else:
            path = files.get_config_dir_path()

        self.lib_path = path / filename
        if self.lib_path.is_file():
            self._xml = xml_load_from_path(self.lib_path)
            self.build_element_dict()
        else:
            logging.warning(f'The library {self.lib_path} does not exist.')

    def __str__(self):
        return f'Library({self.lib_path.name})'

    def build_element_dict(self):
        for child in self._xml:
            name = child.get('name')
            if name:
                child.attrib.pop('name')
                self._elem_dict[name] = child
            else:
                logging.warning('Library element has no name. Skipping.')

    def get_elem(self, name):
        if name in self._elem_dict:
            return self._elem_dict[name]
        else:
            return None




from librec_auto.core.cmd import Cmd
from librec_auto.core.util import Files
from librec_auto.core.util.utils import create_param_spec, get_script_path
from librec_auto.core.util.xml_utils import single_xpath
from librec_auto.core import ConfigCmd
import sys
import subprocess
from lxml import etree
import logging


class RerankCmd(Cmd):

    def __init__(self):
        self._files = None
        self._config = None

    def __str__(self):
        return f'RerankCmd()'

    def setup(self, args):
        pass

    def dry_run(self, config):
        self._files = config.get_files()
        self._config = config

        files = config.get_files()
        if files.get_sub_count() > 0:
            for i in range(0, files.get_sub_count()):
                sub_path = files.get_sub_paths(i)
                script_elem = single_xpath(sub_path.get_exp_conf(), '/librec-auto/rerank/script')
                param_spec = create_param_spec(script_elem)
                script_path = get_script_path(script_elem, 'rerank')
                ref_path = sub_path.get_ref_exp_name()

                print(f'librec-auto (DR): Running re-ranking command {self} for {sub_path.subexp_name}')
                print(f'\tRe-rank script: {script_path}')
                print(f'\tParameters: {param_spec}')
                if ref_path:
                    print(f'\tResults from: {ref_path}')

    # TODO RB 2019-12-12 You can only have one re-ranking script. Others will be silently ignored. Should probably
    # have a warning for this.
    def execute(self, config: ConfigCmd):
        self.status = Cmd.STATUS_INPROC
        self._files = config.get_files()
        self._config = config

        if self._files.get_sub_count() > 0:
            for i in range(0, self._files.get_sub_count()):
                sub_path = self._files.get_sub_paths(i)
                self.rerank(sub_path)

    def rerank(self, sub_path):
        conf_xml = sub_path.get_exp_conf()
        script_elems = conf_xml.xpath('/librec-auto/rerank/script')
        if len(script_elems) > 0:
            script_elem = script_elems[0]
            param_spec = create_param_spec(script_elem)
            script_path = get_script_path(script_elem, 'rerank')
            result_path = sub_path.get_path('result')

            if not result_path.exists():
                print('librec-auto: No results. Cannot rerank ', self._config.get_target())
                return

            if not sub_path.get_ref_exp_name():            # If there is no ref, then LibRec was run
                sub_path.results2original()           # and there are results here
                original_path = sub_path.get_path('original')
            else:                                           # If there is a ref, the results are in
                ref_sub_path = sub_path.get_ref_sub_path()   # that experiment's directory
                original_path = ref_sub_path.get_path('original')

            print(f'librec-auto: Running re-ranking script {script_path} for {sub_path.subexp_name}')
            self.run_script(script_path, sub_path, original_path, param_spec)

            if len(script_elems) > 1:
                logging.warning('More than one re-ranking script found. Ignoring.')


    def run_script(self, script, sub_paths, original_path, param_spec):
        # TODO: RB If rerank only, then leave original folder alone and delete results files
        # TODO: RB If config file is in a non-standard place, I think this will fail. Maybe the sub-scripts
        # should get passed the same information as the original script when invoked?

        proc_spec = [sys.executable, script.as_posix(), self._config.get_files().get_config_path().name,
                     self._config.get_target(),
                     original_path.absolute().as_posix(),
                     sub_paths.get_path('result').absolute().as_posix()] + param_spec
        print("    Parameters: " + str(proc_spec))
        subprocess.call(proc_spec)



from lxml import etree
from pathlib import Path
import copy

def build_parent_path(elem, pathsofar=''):
    if elem.tag == 'param':                     # params are distinguished by name attributes
        pathsofar = f"[@name='{elem.get('name')}']"
    nextpath = '/' + elem.tag + pathsofar
    if (elem.getparent() == None):
        return nextpath
    elif (type(elem) is not etree._Element):
        return nextpath
    else:
        return build_parent_path(elem.getparent(), nextpath)

def read_xml_from_path_string(path_str):
    path = Path(path_str)
    return xml_load_from_path(path)


def xml_load_from_path(path):
    """
    Loads the XML file in a dictionary

    Prints a warning and returns an empty dictionary if the file can't be read.
    :param path: The file name
    :return: A dictionary with the XML rules
     """
    try:
        with path.open() as fd:
            txt = fd.read()
    except IOError as e:
        print ("Error reading ", path)
        print ("IO error({0}): {1}".format(e.errno, e.strerror))
        # logging.error("Error reading %s. IO error: (%d) %s", path, e.errno, e.strerror)
        return {}

    return xml_load_from_text(txt)


def xml_load_from_text(txt):
    try:
        parser = etree.XMLParser(remove_comments=True)
        xml_data = etree.fromstring(txt, parser=parser)
    except etree.XMLSyntaxError as e:
        print ("Error parsing XML")
        print ("LXML error in line: {0}".format(e.lineno))
        xml_data = None

    return xml_data

# The idea is that we have a complete element lib_elem and a partial element mod_elem.
# The mod_elem contents override those in lib_elem when they overlap.
# Only works at the top-level of children in the element.
def merge_elements(lib_elem, mod_elem):
    lib = copy.deepcopy(lib_elem)
    mod = copy.deepcopy(mod_elem)

    lib_tags = [child.tag for child in lib.iterchildren(tag=etree.Element)]
    mod_tags = [child.tag for child in mod.iterchildren(tag=etree.Element)]

    mod_extra = list(set(mod_tags).difference(set(lib_tags)))

    for lchild in lib.iterchildren(tag=etree.Element):
        mchild_lst = [child for child in mod.iterchildren(lchild.tag)]
        if len(mchild_lst) > 0:
            lchild.clear()
            lchild.getparent().replace(lchild, mchild_lst[0])

    for extra_tag in mod_extra:
        matching = [child for child in mod.iterchildren(tag=extra_tag)]
        lib.append(matching[0])

    return lib

def single_xpath(elem, path):
    ans = elem.xpath(path)
    if len(ans) > 0:
        return ans[0]
    else:
        return None




from librec_auto.core.cmd import Cmd
from multiprocessing.dummy import Pool as ThreadPool
from librec_auto.core.util import Status


class ParallelCmd(Cmd):

    def __init__(self, cmds, threads):
        self._thread_count = threads
        if len(cmds) > 0:
            self.set_commands(cmds)
            self.status = Cmd.STATUS_CONFIG
        else:
            self._commands = []
        self._pool = None

    def __str__(self):
        return f'ParallelCmd({len(self._commands)} commands, {self._thread_count} threads)'

    def setup (self, args):
        for cmd in self._commands:
            cmd.setup(args)

    def set_commands(self, cmds):
        self._commands = cmds
        self.status = Cmd.STATUS_CONFIG

    def add_command(self, cmd):
        self._commands.append(cmd)

    def get_commands(self):
        return self._commands

    def dry_run(self, config):
        print (f"librec-auto (DR): Executing parallel, command {self}")
        self.status = Cmd.STATUS_INPROC
        self._pool = ThreadPool(self._thread_count)

        self._pool.map(lambda cmd: cmd.dry_run(config), self._commands)

        self._pool.close()
        self._pool.join()

        for cmd in self._commands:
            if cmd.status == Cmd.STATUS_ERROR:
                self.status = Cmd.STATUS_ERROR
                return
        self.status = Cmd.STATUS_COMPLETE

    def execute (self, config):
        self.status = Cmd.STATUS_INPROC
        self._pool = ThreadPool(self._thread_count)

        self._pool.map(lambda cmd: cmd.execute(config), self._commands)

        self._pool.close()
        self._pool.join()

        for cmd in self._commands:
            if cmd.status == Cmd.STATUS_ERROR:
                self.status = Cmd.STATUS_ERROR
                return
        self.status = Cmd.STATUS_COMPLETE




from abc import ABC, abstractmethod


class Cmd(ABC):

    STATUS_INIT = 0
    STATUS_CONFIG = 1
    STATUS_INPROC = 2
    STATUS_COMPLETE = 3
    STATUS_ERROR = -1

    def __init__(self):
        self.status = self.STATUS_INIT

    def __repr__(self):
        return "Cmd()"

    def __str__(self):
        return f"Cmd({type(self)})"

    @abstractmethod
    def setup (self, args):
        pass

    @abstractmethod
    def dry_run(self, config):
        print (f"librec-auto (DR): Executing command {self}")

    @abstractmethod
    def execute (self, config):
        pass

class SequenceCmd(Cmd):

    _commands = []

    def __str__(self):
        return f'SequenceCmd({len(self._commands)} commands)'

    def __init__(self, cmds):
        if len(cmds) > 0:
            self.set_commands(cmds)

    def setup (self, args):
        for cmd in self.commands:
            cmd.setup(args)

    def set_commands(self, cmds):
        self._commands = cmds
        self.status = Cmd.STATUS_CONFIG

    def add_command(self, cmd):
        self._commands.append(cmd)

    def get_commands(self):
        return self._commands

    def dry_run(self, config):
        print (f"librec-auto (DR): Executing sequence command {self}")
        for cmd in self._commands:
            cmd.dry_run(config)

    def execute (self, config):
        self.status = Cmd.STATUS_INPROC
        for cmd in self._commands:
            cmd.execute(config)
            if cmd.status == Cmd.STATUS_ERROR:
                self.status = Cmd.STATUS_ERROR
                return
        self.status = Cmd.STATUS_COMPLETE



import itertools
from collections import defaultdict
import copy

# type_factor stores the product of the counts of the values for that type of variable.
# The product of all of the type factors is the total number of experiments. The
# type_factor for the "librec" type gives the number of calls to LibRec.
class VarColl:
    def __init__(self):
        self.vars = defaultdict(list)
        self.var_confs = []
        self.types = []
        self.type_factor = defaultdict(lambda: 1)

    def compute_var_configurations(self):
        # Need to reverse so that the cartesian product has the right priority
        rev_types = copy.copy(self.types)
        rev_types.reverse()
        configs = [self.get_type_conf(type) for type in rev_types]
        conf_product = list(itertools.product(*configs))
        self.var_confs = [self.conf_merge(conf_elem) for conf_elem in conf_product]
        self.set_refs()

    def add_var(self, type, path, vals):
        if len(self.vars[type]) == 0:
            self.types.append(type)
        self.vars[type].append(VarInfo(type, path, vals))
        self.type_factor[type] = self.type_factor[type] * len(vals)

    def get_vars(self, type):
        return self.vars[type]

    def get_type_conf(self, type):
        vars = self.get_vars(type)
        flat_vars = [var.flatten() for var in vars]
        flat_vars.reverse()
        var_confs = list(itertools.product(*flat_vars))
        return [VarConfig(type, var_conf) for var_conf in var_confs]

    def conf_merge(self, confs):
        if len(confs) == 1:
            confs[0].type = 'merged'
            return confs[0]
        base = copy.deepcopy(confs[0])
        for conf in confs[1:]:
            base.merge(conf)
        base.type = 'merged'
        # print(str(base.vars))
        return base

    def set_refs(self):
        librec_factor = self.type_factor['librec']
        for i in range(0, len(self.var_confs)):
            row = i // librec_factor
            col = i % librec_factor
            if row > 0:
                self.var_confs[i].ref_config = self.var_confs[col]


class VarConfig:
    def __init__(self, type, vars=()):
        self.exp_no = None
        self.exp_dir = None
        self.type = type
        self.vars = vars
        self.ref_config = None

    def __str__(self):
        return f'<VarConfig {self.type} Vars: {self.vars}'

    def merge(self, vconf):
        self.vars = vconf.vars + self.vars

class VarInfo:

    def __init__(self, type, path, vals):
        self.type = type
        self.path = path
        self.val = vals

    def __str__(self):
        return f'<VarInfo {self.type} {self.path} V: {self.val}>'

    def __repr__(self): return self.__str__()

    def flatten(self):
        if type(self.val) is list:
            return [VarInfo(self.type, self.path, val) for val in self.val]
        else:
            return self




from lxml import etree
from inspect import getsourcefile
from os.path import abspath
from pathlib import Path
import logging
from . import xml_utils

# TODO: Replace with XPath
def safe_xml_path(config, key_list):
    """
    Checks that the list of keys in key_list can be used to navigate through
    the dictionaries in config.

    :param config: a nested dictionary structure
    :param path_list: a list of dictionary keys
    :return: True if the sequence of keys is valid
    """
    dct = config
    for elem in key_list:
        if elem in dct:
            dct = dct[elem]
        else:
            return False
    return True

# TODO: Replace with XPath
def extract_from_path(config, key_list):
    """
    Retrieves element content by following the path in key_list.

    :param config: a nested dictionary structure
    :param path_list: a list of dictionary keys
    :return: Element at the end of the sequence
    """
    dct = config
    for elem in key_list:
        if elem in dct:
            dct = dct[elem]
        else:
            return None
    return dct


def force_list(item):
    """
    Ensures that an item is of list type.

    :param item:
    :return: item or [item]
    """
    if (type(item) is list):
        return item
    else:
        return [item]

def force_path(item):
    """
    Ensures that an item is of Path type.

    :param item:
    :return: item or [item]
    """
    if (type(item) is Path):
        return item
    else:
        return Path(item)


def frange(start, stop, step):
     x = start
     while x < stop:
         yield x
         x += step


def confirm(prompt=None, resp=False):
    """prompts for yes or no response from the user. Returns True for yes and
    False for no.

    'resp' should be set to the default value assumed by the caller when
    user simply types ENTER.

    >>> confirm(prompt='Create Directory?', resp=True)
    Create Directory? [y]|n:
    True
    >>> confirm(prompt='Create Directory?', resp=False)
    Create Directory? [n]|y:
    False
    >>> confirm(prompt='Create Directory?', resp=False)
    Create Directory? [n]|y: y
    True

    """

    if prompt is None:
        prompt = 'Confirm'

    if resp:
        prompt = '%s [%s]|%s: ' % (prompt, 'y', 'n')
    else:
        prompt = '%s [%s]|%s: ' % (prompt, 'n', 'y')

    while True:
        # ans = raw_input(prompt)
        ans = input(prompt)
        if not ans:
            return resp
        if ans not in ['y', 'Y', 'n', 'N']:
            print ('please enter y or n.')
            continue
        if ans == 'y' or ans == 'Y':
            return True
        if ans == 'n' or ans == 'N':
            return False


def get_script_path(script_xml, cmd_type):
    script_path = '.'
    if script_xml.get('lang') != 'python3':
        print(f'librec-auto: Only Python3 scripts currently supported. Got {script_xml.get("lang")}.')
        return None
    if script_xml.get('src'):
        if script_xml.get('src') == 'system':
            script_path = Path(abspath(getsourcefile(lambda:0))).parent.parent / 'cmd' / cmd_type
        else:
            script_path = force_path(script_xml.get('src'))
    name_elem = xml_utils.single_xpath(script_xml, 'script-name')
    if name_elem is not None:
        return script_path / name_elem.text
    else:
        return None

def create_param_spec(script_xml):
    params = script_xml.xpath('param')
    param_list = []
    for param in params:
        key = param.get('name')
        val = param.text
        param_list.append(f'--{key}={val}')
    return param_list




from pathlib import Path
import hashlib
import inspect
import librec_auto
from librec_auto.core.util.utils import force_path
from librec_auto.core.util.xml_utils import xml_load_from_path
from collections import OrderedDict
import glob
import shutil
import logging
import re

class Files:
    """
    This class encapsulates the file system for librec-auto.

    The files for librec-auto can be stored in a number of possible locations:
    - global location: determined by the install location of librec-auto
    - rules-specific location: determined by the location of a rules directory [Not implemented]
    - user-specific location: associated with the user's home directory [Not implemented]
    - experiment-specific location: in a directory associated with a specific experiment.

    These are the types of files that are managed:
    - configuration files (XML). Stores information about how experiments are configured.
    - properties files (key=value format). Generated from configuration files for input to librec
    - split files (UIR or UIRT format). Generated by separating input rules files into different test/training splits
    - experiment logs (csv files). Content is algorithm-specific. Generated by a run of an experiment

    NOTE: User directory is currently unused.
    """

    _DEFAULT_GLOBAL_DIR_STR = inspect.getfile(librec_auto)

    _DEFAULT_CONFIG_DIR_NAME = "conf"
    _DEFAULT_RULES_DIR_NAME = "rules"
    _DEFAULT_LIB_DIR_NAME = "librec_auto/library"
    _DEFAULT_RES_DIR_NAME = "result"
    _DEFAULT_SPLIT_DIR_NAME = "data/split"
    _DEFAULT_JAR_DIR_NAME = "librec_auto/jar"
    _DEFAULT_POST_DIR_NAME = "post"
    _DEFAULT_LIBRARY_DIR_NAME = "lib"
    _EXP_DIR_FORMAT = "exp{:05d}"
    _EXP_DIR_PATTERN = "exp(\d+)"

    DEFAULT_PROP_FILE_NAME = "librec.properties"
    _DEFAULT_LA_JAR = "auto.jar"
    _DEFAULT_LR_JAR = "librec.jar"
    _DEFAULT_RULES_FILE = "librec_auto/rules/element-rules.xml"

    DEFAULT_CONFIG_FILENAME = "config.xml"
    DEFAULT_REF_EXP_FILENAME = "ref_exp.txt"

    def __init__(self):
        self._config_dir_path = Path(self._DEFAULT_CONFIG_DIR_NAME)
        self._split_dir_path = Path(self._DEFAULT_SPLIT_DIR_NAME)
        self._jar_dir_path = Path(self._DEFAULT_JAR_DIR_NAME)
        self._post_dir_path = Path(self._DEFAULT_POST_DIR_NAME)
        self._lib_dir_path = Path(self._DEFAULT_LIBRARY_DIR_NAME)
        self._sub_path_dict = OrderedDict()

        module_init_path = Path(Files._DEFAULT_GLOBAL_DIR_STR).parent
        self._global_path = module_init_path.parent

    def get_global_path(self): return self._global_path

    def get_jar_path(self): return self.get_global_path() / self._jar_dir_path

    def get_exp_path(self): return self._exp_path

    def get_split_path(self): return self._exp_path / self._split_dir_path

    def get_config_path(self): return self._exp_path / self._config_dir_path / self._config_file_name

    def get_config_dir_path(self): return self._exp_path / self._config_dir_path

    def get_post_path(self): return self._exp_path / self._post_dir_path

    def set_global_path(self, path): self._global_path = Path(path)

    def set_config_dir_path(self, path): self._config_dir_path = Path(path)

    def set_exp_path(self, path): self._exp_path = Path(path)

    def set_config_file(self, filename): self._config_file_name = Path(filename)

    def get_rules_path (self):
        return self.get_global_path() / self._DEFAULT_RULES_FILE

    def get_lib_path (self):
        return self.get_global_path() / self._DEFAULT_LIB_DIR_NAME

    # 2019-11-23 RB TODO: Separate librec.jar and auto.jar files. Then restore the two jar classpaths here.
    def get_classpath (self):
        # return (self.get_global_path() / self.get_jar_path() / self._DEFAULT_LA_JAR).absolute().as_posix() + ";" + \
        #        (self.get_global_path() / self.get_jar_path() / self._DEFAULT_LR_JAR).absolute().as_posix()
        return (self.get_jar_path() / self._DEFAULT_LA_JAR).absolute().as_posix()

    def get_subexp_name (self, count):
        return self._EXP_DIR_FORMAT.format(count)

    def detect_sub_path(self, exp_no):
        return (self.get_exp_path() / self.get_subexp_name(exp_no)).exists()

    def detect_sub_paths (self, count=0):
        sub_count = 0
        while self.detect_sub_path(sub_count):
            self._sub_path_dict[sub_count] = SubPaths(self, self.get_subexp_name(sub_count), create=False)
            sub_count += 1
        if sub_count != count:
            print(f'librec-auto: Expecting {count} existing experiment directories in {self.get_exp_path()}. Found {sub_count}.')

    def create_sub_paths (self, tuple_count):
        if tuple_count == 0:
            sub_exp_count = 1
        else:
            sub_exp_count = tuple_count

        for i in range(sub_exp_count):
            self._sub_path_dict[i] = SubPaths(self, self.get_subexp_name(i), create=True)

    def ensure_sub_paths(self, exp_count):
        if self.detect_sub_path(0):
            self.detect_sub_paths(count=exp_count)
        else:
            self.create_sub_paths(exp_count)

    def get_sub_count (self):
        if len(self._sub_path_dict) > 0:
            return max(self._sub_path_dict.keys())+1
        else:
            return 0

    def get_sub_paths(self, count):
        if count in self._sub_path_dict:
            return self._sub_path_dict[count]
        else:
            return None

    def get_sub_paths_by_name(self, name):
        for subp in self._sub_path_dict.values():
            if subp.subexp_name == name:
                return subp
        return None

    def get_sub_paths_iterator(self):
        return self._sub_path_dict.values()

    @staticmethod
    def dir_hash(maybe_path):
        """
        Starting from a directory, gets all subdirectories, extracts the individual files and creates a hash value
        using the file name, size, and last modification date.

        Probably just modification date is enough.
        :param maybe_path:
        :return:
        """
        hasher = hashlib.sha1()
        path = force_path(maybe_path)
        full_listing = path.glob('**/*')
        files = [fl for fl in full_listing if fl.is_file()]
        for fl in files:
            fl_stat = fl.stat()
            fl_size = fl_stat.st_size
            fl_date = fl_stat.st_mtime
            fl_info = "{}-{}-{}".format(fl.name, fl_size, fl_date)
            fl_bytes = fl_info.encode('utf-8')
            hasher.update(fl_bytes)
        return hasher.hexdigest()

class SubPaths:
    """
    Represents the various directories and paths associated with a single sub-experiment

    - log
    - result
    - original (for re-ranking)
    - conf
    """
    DEFAULT_LOG_FILENAME = 'librec.log'

    _prop_dict = {'log': 'dfs.log.dir',
#                 'split': 'dfs.split.dir',
                 'result': 'dfs.result.dir',
                  'conf': 'dfs.config.dir'}

    _sub_dirs = ['conf', 'log', 'result', 'original']

    def __init__(self, files, subexp_name, create=True):
        self._path_dict = {}
        self.subexp_name = subexp_name
        self.files = files

        subexp_path = files.get_exp_path() / subexp_name
        self.set_path('subexp', subexp_path)

        status_path = subexp_path / '.status'
        self.set_path('status', status_path)

        for subdir in self._sub_dirs:
            subdir_path = subexp_path / subdir
            self.set_path(subdir, subdir_path)

        if create:
            logging.info("Creating subexperiment: {}", subexp_name)
            subexp_path.mkdir(exist_ok=True)
            for subdir in self._sub_dirs:
                self.get_path(subdir).mkdir(exist_ok=True)

    def get_path(self, type):
        if type in self._path_dict:
            return self._path_dict[type]
        else:
            return None

    def get_librec_properties_path(self):
        return self.get_path('conf') / Files.DEFAULT_PROP_FILE_NAME

    def get_ref_exp_flag_path(self):
        return self.get_path('conf') / Files.DEFAULT_REF_EXP_FILENAME

    def get_path_str(self, type):
        return self.get_path(type).as_posix()

    def get_path_platform(self, type):
        return str(self.get_path(type))

    def get_path_prop(self, type):
        return self._prop_dict[type]

    def set_path(self, type, path):
        self._path_dict[type] = path

    def set_path_from_string(self, type, path_str):
        self._path_dict[type] = Path(path_str)

    def get_exp_conf(self):
        path = self.get_path('conf') / Files.DEFAULT_CONFIG_FILENAME
        xml_input = xml_load_from_path(path)
        return xml_input

    def get_ref_exp_name(self):
        ref_flag_file = self.get_ref_exp_flag_path()
        if ref_flag_file.exists():
            with ref_flag_file.open() as fh:
                exp_name = fh.readline()
                return exp_name.rstrip()
        else:
            return None

    def get_ref_sub_path(self):
        ref = self.get_ref_exp_name()
        if not ref:
            return None
        else:
            return self.files.get_sub_paths_by_name(ref)

    # Assumes path is set up
    def add_to_config(self, config, type):
        prop_name = self.get_path_prop(type)
        prop_val = self.get_path_str(type)
        config[prop_name] = prop_val

    def results2original(self):
        original_path = self.get_path('original')
        result_path = self.get_path('result')
        shutil.rmtree(original_path)
        original_path.mkdir()
        files = glob.glob((result_path / '*').as_posix())
        for file in files:
            shutil.copy2(file, original_path)

    def original2results(self):
        original_path = self.get_path('original')
        result_path = self.get_path('result')
        files = glob.glob((original_path / '*').as_posix())
        for file in files:
            shutil.copy2(file, result_path)
        shutil.rmtree(original_path)



from .files import Files, SubPaths
from .utils import safe_xml_path, extract_from_path, force_list, force_path, frange, confirm
from .xml_utils import read_xml_from_path_string, xml_load_from_path, xml_load_from_text, \
    build_parent_path, merge_elements, single_xpath
from .log_file import LogFile
from .status import Status
from .librec_properties import LibrecProperties
from .library import Library, LibraryColl
from .var import VarColl, VarInfo, VarConfig


from collections import OrderedDict
from librec_auto.core.util import Files, utils
import logging
import itertools
from pathlib import Path

class ConfigLib:

    _xml = None
    _element_dict = None
    _source_path = None

    def read_xml(self, path_str):
        path = Path(path_str)
        if (path.exists()):
            xml_input = utils.xml_load_from_path(path)
            return xml_input
        else:
            return None

    def __init__(self, path):
        self._source_path = path
        self._xml = self.read_xml(path)
        self.build_element_dict()

    def build_element_dict(self):
        self._element_dict = {}
        xml_contents = self._xml['librec-auto-library']
        for elem_type in xml_contents.keys():
            elems = utils.force_list(xml_contents[elem_type])
            for elem in elems:
                if '@name' in elem:
                    name = elem['@name']
                    if name not in self._element_dict:
                        self._element_dict[name] = elem
                    else:
                        print(f'librec-auto: WARNING: Name {name} is not unique in {self._source_path}')
                else:
                    print(f'librec-auto: WARNING: Element {elem} has no name in {self._source_path}')

    def exists_element(self, ref):
        return ref in self._element_dict

    def get_element(self, ref):
        return self._element_dict[ref]

# Library are searched in inverse order from when they are added. Most recent have priority.
class ConfigLibCollection:

    _libs = []

    def __init__(self):
        pass

    def append(self, lib):
        self._libs = [lib] + self._libs

    def exists_element(self, ref):
        return any([lib.exists_element() for lib in self._libs])

    def get_element(self, ref):
        for lib in self._libs:
            if lib.exists_element(ref):
                return lib.get_element(ref)
        # Should probably throw a key not found error
        return None






from librec_auto.core.cmd import Cmd
from librec_auto.core.util import Files
from librec_auto.core import ConfigCmd
import os
import sys
import librec_auto

import progressbar
import urllib.request
import time

class InstallCmd(Cmd):

    def __str__(self):
        return f'InstallCmd()'

    def setup(self, args):
        pass

    def dry_run(self, config):
        # self._config = config
        print (f'librec-auto (DR): Running install command {self}')
        # for script_path, params in config.collect_scripts('post'):
        #     param_spec = utils.create_param_spec(params)
        #     print (f'    Post script: {script_path}')
        #     print (f'\tParameters: {param_spec}')


    def execute(self, config: ConfigCmd):
        jar_path = os.path.dirname(librec_auto.__file__) + '\\jar\\auto.jar'
        import urllib.request
        urllib.request.urlretrieve('https://www.dropbox.com/s/hyemqt99790t16q/auto.jar?dl=1', jar_path, self.show_progress)
    # def execute(self, config: ConfigCmd):
    #     lib_path = os.path.dirname(librec_auto.__file__)
    #     jar_path = lib_path.partition("\\librec_auto")[0]
    #     jar_path += '\\jar\\auto.jar'
    #     import urllib.request
    #     urllib.request.urlretrieve('https://www.dropbox.com/s/hyemqt99790t16q/auto.jar?dl=1', jar_path, self.show_progress)

    def show_progress(self, count, block_size, total_size):
        # global start_time
        if count == 0:
            self.start_time = time.time()
            return
        duration = time.time() - self.start_time
        progress_size = int(count * block_size)
        speed = int(progress_size / (1024 * (int(duration) + 1))) # int(progress_size / (1024 * duration))
        percent = min(int(count*block_size*100/total_size),100) # int(count * block_size * 100 / total_size)
        sys.stdout.write("\r...%d%%, %d MB, %d KB/s, %d seconds passed" %
                         (percent, progress_size / (1024 * 1024), speed, duration))
        sys.stdout.flush()

    # def show_progress(self, block_num, block_size, total_size):
    #     # global pbar
    #     if self.pbar is None:
    #         self.pbar = progressbar.ProgressBar(maxval=total_size)
    #
    #     downloaded = block_num * block_size
    #     if downloaded < total_size:
    #         self.pbar.update(downloaded)
    #     else:
    #         self.pbar.finish()
    #         self.pbar = None


# class MyProgressBar():
#     def __init__(self):
#         self.pbar = None
#
#     def __call__(self, block_num, block_size, total_size):
#         if not self.pbar:
#             self.pbar=progressbar.ProgressBar(maxval=total_size)
#             self.pbar.start()
#
#         downloaded = block_num * block_size
#         if downloaded < total_size:
#             self.pbar.update(downloaded)
#         else:
#             self.pbar.finish()


import argparse
from pathlib import Path
from librec_auto.core import read_config_file
from librec_auto.core.util import Files
from librec_auto.core.cmd import Cmd, SetupCmd, SequenceCmd, PurgeCmd, LibrecCmd, PostCmd, RerankCmd, StatusCmd, ParallelCmd, InstallCmd
import logging


def read_args():
    '''
    Parse command line arguments.
    :return:
    '''
    parser = argparse.ArgumentParser(description='The librec-auto tool for running recommender systems experiments')
    parser.add_argument('action',choices=['run','split', 'eval', 'rerank', 'post', 'purge', 'status', 'describe', 'check', 'install'])

    parser.add_argument("target", help="Path to experiment directory")

    # Optional with arguments
    # parser.add_argument("-ex","--exhome", help="stub")
    # parser.add_argument("-rs","--reset", help = "stub")
    # parser.add_argument("-rss","--revise-step", help="stub")
    parser.add_argument("-c", "--conf", help="Use the specified configuration file")

    # Flags
    parser.add_argument("-dr","--dry_run",
                        help = "Show sequence of command execution but do not execute commands",
                        action ="store_true")
    parser.add_argument("-q","--quiet",
                        help = "Skip confirmation when purging",action ="store_true")
    parser.add_argument("-np","--no_parallel",
                        help = "Ignore thread-count directive and run all operations sequentially",
                        action ="store_true")
    parser.add_argument("-p", "--purge",
                        help="Purge results of step given in <option> and all subsequent steps",
                        choices=['all', 'split', 'results', 'rerank', 'post'], default='all')
    parser.add_argument("-nc", "--no_cache",
                        help="Do not cache any intermediate results (Not implemented)",
                        action="store_true")

    input_args = parser.parse_args()
    return vars(input_args)


def load_config(args):

    config_file =  Files.DEFAULT_CONFIG_FILENAME

    if args['conf']:      # User requested a different configuration file
        config_file = args['conf']

    target = args['target']

    return read_config_file(config_file, target)


DESCRIBE_TEXT = 'Librec-auto automates recommender systems experimentation using the LibRec Java library.\n' +\
    '\tA librec-auto experiment consist of five steps governed by the specifications in the configuration file:\n' +\
    '\t- split: Create training / test splits from a data set. (LibRec)\n'+\
    '\t- exp: Run an experiment generating recommendations for a test set (LibRec)\n' +\
    '\t- rerank (optional): Re-rank the results of the experiment (script)\n' +\
    '\t- eval: Evaluate the results of a recommendation experiment (LibRec)\n' +\
    '\t- post (optional): Perform post-processing computations (script)\n' + \
    'Steps labeled LibRec are performed by the LibRec library using configuration properties generated by librec-auto.\n' +\
    'Steps labeled script are performed by experimenter-defined scripts.\n' + \
    'Run librec_auto describe <step> for additional information about each option.'

DESCRIBE_DICT = {
    'run': 'Run a complete librec-auto experiment. Re-uses cached results if any. \
May result in no action if all computations are up-to-date and no purge option is specified.',
    'split': 'Run the training / test split only',
    'exp': 'Run the experiment, re-ranking, evaluation, and post-processing',
    'rerank': 'Run the re-ranking, evaluation and post-processing',
    'eval' : 'Run the evaluation and post-processing',
    'post': 'Run post-processing steps',
    'purge': 'Purge cached computations. Uses -p flag to determine what to purge',
    'status': 'Print out the status of the experiments'
}


def print_description(args):
    act = args['target']
    if act in DESCRIBE_DICT:
        print (f'core {act} <target>: {DESCRIBE_DICT[act]}')
    else:
        print(DESCRIBE_TEXT)


def purge_type (args):
    if 'purge' in args:
        return args['purge']
    # If no type specified and you're purging, purge everything
    elif args['action']=='purge':
        return 'split'
    else:
        return 'none'

# TODO: Need to rewrite as "build_exec_commands" where the action incorporates both execution
# and reranking. Remember that the re-ranker only requires one run of the prediction algorithm for any
# variation its own parameters.
def build_librec_commands(librec_action, args, config):
    librec_commands = [LibrecCmd(librec_action, i) for i in range(config.get_sub_exp_count())]
    threads = config.thread_count()

    if threads > 1 and not args['no_parallel']:
        return ParallelCmd(librec_commands, threads)
    else:
        return SequenceCmd(librec_commands)

# The purge rule is: if the command says to run step X, purge the results of X and everything after.
def setup_commands (args, config):
    action = args['action']
    purge_noask = args['quiet']

    # Create flags for optional steps
    rerank_flag = config.has_rerank()
    post_flag = config.has_post()

    # Purge files (possibly) from splits and subexperiments
    if action == 'purge':
        cmd = PurgeCmd(purge_type(args), noask=purge_noask)
        return cmd

    # Shows the status of the experiment
    if action == 'status':
        cmd = StatusCmd()
        return cmd

    # Perform (only) post-processing on results
    if action == 'post' and post_flag:
        cmd = PostCmd()
        return cmd
    # No post scripts available
    if action == 'post' and not post_flag:
        logging.warning("No post-processing scripts available for post command.")
        return None

    # Perform re-ranking on results, followed by evaluation and post-processing
    if action == 'rerank' and rerank_flag: # Runs a reranking script on the python side
        cmd1 = PurgeCmd('rerank', noask=purge_noask)
        cmd2 = SetupCmd()
        cmd3 = RerankCmd()
        cmd4 = build_librec_commands('eval', args, config)
        cmd = SequenceCmd([cmd1, cmd2, cmd3, cmd4])
        if post_flag:
            cmd.add_command(PostCmd())
        return cmd
    # No re-ranker available
    if action == 'rerank' and not rerank_flag:
        logging.warning("No re-ranker available for rerank command.")
        return None

    # LibRec actions
    # re-run splits only
    if action == 'split':
        cmd1 = PurgeCmd('split', noask=purge_noask)
        cmd2 = SetupCmd()
        cmd3 = build_librec_commands('split', args, config)
        cmd = SequenceCmd([cmd1, cmd2, cmd3])
        return cmd

    # re-run experiment and continue
    if action== 'run':
        cmd1 = PurgeCmd('results', noask=purge_noask)
        cmd2 = SetupCmd()
        cmd3 = build_librec_commands('full', args, config)
        cmd = SequenceCmd([cmd1, cmd2, cmd3])
        if rerank_flag:
            cmd.add_command(RerankCmd())
            cmd.add_command(build_librec_commands('eval', args, config))
        if post_flag:
            cmd.add_command(PostCmd())
        return cmd

    # eval-only
    if action == 'eval':
        cmd1 = PurgeCmd('post', noask=purge_noask)
        cmd2 = SetupCmd()
        cmd3 = build_librec_commands('eval', args, config)
        cmd = SequenceCmd([cmd1, cmd2, cmd3])
        if post_flag:
            cmd.add_command(PostCmd())
        return cmd

    if action == 'check':
        cmd = build_librec_commands('check', args, config)
        return cmd

    if action == 'install':
        cmd = InstallCmd()
        return cmd

# -------------------------------------


if __name__ == '__main__':
    args = read_args()

    if args['action']=='describe':
        print_description(args)
    else:
        config = load_config(args)

        if config.is_valid():
            command = setup_commands(args, config)
            if isinstance(command, Cmd):
                if args['dry_run']:
                    command.dry_run(config)
                else:
                    command.execute(config)
            else:
                logging.error("Command instantiation failed.")
        else:
            logging.error("Configuration loading failed.")



from collections import OrderedDict, defaultdict
from librec_auto.core.util import Files, utils, build_parent_path, xml_load_from_path, SubPaths
from lxml import etree
import copy
import logging
import itertools

# 2020-06-25 RB All new implementation


class LibrecTranslation:

    __instance = None

    def __new__(cls, files):
        if LibrecTranslation.__instance is None:
            LibrecTranslation.__instance = object.__new__(cls)
            LibrecTranslation.files = files
        return LibrecTranslation.__instance

    def __init__(self, files):
        self.xml = self.read_rules(files)

    def read_rules(self, files):
        rules_path = self.files.get_rules_path()
        if (rules_path.exists()):
            rules_input = xml_load_from_path(rules_path)
            return rules_input
        else:
            logging.warning(f'No translation rules read from {rules_path}.')
            return None

class LibrecProperties:

    def __init__(self, xml, files):
        self.properties = {}
        self._conf_xml = xml
        self.process_config(files)
        self.add_property_hacks()

    def add_property_hacks(self):

        self.properties['dfs.split.dir'] = "split"

    def process_config(self, files):
        trans_rules = LibrecTranslation(files)
        if type(self._conf_xml) is etree._Element:
            if type(trans_rules.xml) is etree._Element:
                self.process_aux(self._conf_xml, trans_rules.xml)
            else:
                logging.error("Element translations missing.")
        else:
            logging.error(f"Error processing experiment configuration file.")

# ctree: configuration tree
# rtree: rules tree
# recursively walk the ctree and the rtree at the same time
# if an element is labeled in the rules as "no-translate", that means it's contents are not passed to LibRec
#     and it can be ignored
#
    def process_aux(self, conf_tree: etree._Element, rule_tree: etree._Element):
        for conf_elem in conf_tree.iterchildren(tag=etree.Element):
            conf_tag = conf_elem.tag
            rule_elem = rule_tree.findall(conf_tag)
            if len(rule_elem) > 0:                                  # If the entry corresponds to one or more rules
                if len(rule_elem) == 1:
                    rule_elem = rule_elem[0]
                    action_attr = rule_elem.get('action', default=None)
                    if action_attr is not None and action_attr=='no-translate': # If labeled "no translate", skip
                        pass
                    elif len(conf_elem) > 0:                            # If the config file has subelements
                        if len(rule_elem) > 0:                      # If the rules also have subelements
                            self.process_aux(conf_elem, rule_elem)  # recursive call
                        else:                                       # If no corresponding elements in rules
                            logging.warning(f'Mismatch between element contents: {conf_elem} and rule: {rule_elem}')
                    else:                                               # Otherwise, it is a key-value pair
                        self.add_property(rule_elem.text, conf_elem.text)

                else:                           # If the rules have a list,
                    self.process_multi(conf_elem, rule_elem)            # We are distinguishing cases based on attribute

            else: # If the key isn't in the rules, ignore it but warn because it is probably an error.
                logging.warning(f"Tag {conf_tag} is not in element rules.")

    # Two cases:
    # (a) We have multiple matching rules distinguished by attribute. For example,
    # 		<data-file>data.input.path</data-file>
    # 		<data-file format="*">data.model.format</data-file>
    #     In this case, the non-attribute version is handled in the normal way, but a second property is
    #     added, associating the associated attribute value with the key in that rule element.
    # (b) We have multiple matching rules not distinguished by attribute. For example,
    #       <l1-reg>rec.regularization.lambda1</l1-reg>
    #       <l1-reg>rec.slim.regularization.l1</l1-reg>
    #     In this case, we add all of the property mappings since unneeded ones are ignored by LibRec.
    #     The alternative would be to rename these for each algorithm, which is a lot for experimenters to have
    #     to remember.
    # The only difference between these cases is whether the rule element has an attribute or not.
    def process_multi(self, conf_elem, rule_elems):
        for rule_elem in rule_elems:
            key = rule_elem.text
            if rule_elem.attrib:
                attr = rule_elem.keys()[0]                      # Case a
                val = conf_elem.get(attr)
            else:
                val = conf_elem.text
            self.add_property(key, val)


    # Multiple adds mean that the property is list-valued
    def add_property(self, key, val):
        if key in self.properties:
            old_val = utils.force_list(self.properties[key])
            old_val.append(val)
            self.properties[key] = old_val
        else:
            self.properties[key] = val

    def get_property(self, key):
        if key in self.properties:
            val = self.properties[key]
            if type(val) is list:
                return ','.join(val)
            else:
                return val
        else:
            return None

    def get_entries(self):
        return [(key, self.get_property(key)) for key in self.properties.keys()]

    def save(self, exp: SubPaths):
        path = exp.get_path('conf') / Files.DEFAULT_PROP_FILE_NAME
        with path.open('w') as fd:
            for (key, val) in self.get_entries():
                fd.write(f'{key}: {val}\n')





from librec_auto.core.cmd import Cmd
from librec_auto.core.util import Files, SubPaths
from librec_auto.core import ConfigCmd
from  librec_auto.core.util import confirm
import shutil
import os


class SetupCmd (Cmd):

    def __str__(self):
        return 'SetupCmd()'

    def setup(self, args):
        pass

    def dry_run(self, config):
        print(f"librec-auto (DR): Executing setup command {self}")
        config.ensure_experiments()
        config.setup_exp_configs()

    def execute(self, config: ConfigCmd):
        config.ensure_experiments()
        config.setup_exp_configs()




from .cmd import Cmd, SequenceCmd
from .setup_cmd import SetupCmd
from .librec_cmd import LibrecCmd
from .purge_cmd import PurgeCmd
from .post_cmd import PostCmd
from .rerank_cmd import RerankCmd
from .status_cmd import StatusCmd
from .parallel_cmd import ParallelCmd
from .install_cmd import InstallCmd



from librec_auto.core.cmd import Cmd
from librec_auto.core.util import Files, utils
from librec_auto.core import ConfigCmd
from pathlib import Path
import os
import sys
import subprocess


class PostCmd(Cmd):

    POST_SCRIPT_PATH = "core/cmd/post"
    POST_ELEM_XPATH = '/librec-auto/post/script'

    def __init__(self):
        self._config = None

    def __str__(self):
        return f'PostCmd()'

    def setup(self, args):
        pass

    def dry_run(self, config):
        self._config = config
        print (f'librec-auto (DR): Running post command {self}')

        post_elems = config.get_xml().xpath(self.POST_ELEM_XPATH)

        for post_elem in post_elems:
            param_spec = utils.create_param_spec(post_elem)
            script_path = utils.get_script_path(post_elem, 'post')

            print (f'\tPost script: {script_path}')
            print (f'\tParameters: {param_spec}')

    def execute(self, config: ConfigCmd):
        self._config = config
        self.status = Cmd.STATUS_INPROC
        files = config.get_files()

        target = files.get_exp_path()

        post_path = files.get_post_path()

        if not post_path.exists():
            print('librec-auto: post directory missing. Creating. ', target)
            os.makedirs(str(post_path))

        post_elems = config.get_xml().xpath(self.POST_ELEM_XPATH)

        for post_elem in post_elems:
            param_spec = utils.create_param_spec(post_elem)
            script_path = utils.get_script_path(post_elem, 'post')

            proc_spec = [sys.executable, script_path.absolute().as_posix(),
                         self._config.get_files().get_config_path().name,
                         config.get_target()] + param_spec
            print (f'librec-auto: Running post-processing script {proc_spec}')
            subprocess.call(proc_spec)



from collections import OrderedDict, defaultdict
from librec_auto.core.util import Files, utils, build_parent_path, LibrecProperties, \
    xml_load_from_path, Library, LibraryColl, merge_elements, VarColl
from lxml import etree
import copy
import logging
from pathlib import Path
from librec_auto.core.config_lib import ConfigLibCollection, ConfigLib
import re


class ConfigCmd:
    """
    Loads the configuration file, inserts appropriate library contents,
    identifies the parameter variations and creates separate configurations
    for all combinations.
    """

    _PARAM_NAME_PATH_RE = ".+\[@name='(.+)'\]"

    def __init__(self, config_file, target):

        self._files = Files()
        self._target = target

        self._files.set_exp_path(target)
        self._files.set_config_file(config_file)

        self._xml_input = self.read_xml(self._files.get_config_path())
        #self._var_librec_data = defaultdict(list)
        #self._var_rerank_data = defaultdict(list)
        #self._var_params = []
        #self._var_tuples = []
        self._var_coll = VarColl()
        self._libraries = LibraryColl()

    def get_target(self):
        return self._target

    def get_xml(self):
        return self._xml_input

    # def get_var_data(self):
    #     return self._var_data

    def get_value_conf(self, subexp_no):
        return self._var_coll.var_confs[subexp_no]

    def get_sub_exp_count(self):
        exp_count = len(self._var_coll.var_confs)
        if exp_count == 0:
            return 1
        else:
            return exp_count

    def get_files(self):
        return self._files

    def read_xml(self, path_str):
        path = self._files.get_config_path()
        if (path.exists()):
            xml_input = xml_load_from_path(path)
            return xml_input
        else:
            return None

    def ensure_experiments(self):
        exp_count = len(self._var_coll.var_confs)
        if exp_count == 0:
            exp_count = 1
        self.get_files().ensure_sub_paths(exp_count)

    def load_libraries(self):
        lib_paths = []
        lib_elems = self._xml_input.xpath('/librec-auto/library')
        for elem in lib_elems:
            self._libraries.add_lib(Library(elem.text, elem.get('src'), self._files))

    # Process config takes the config file and produces a dictionary of the following form:
    # xpath-string => list of values
    # or xpath-string => (range-to, range-from) pair
    # Right now, we will assume the first
    def process_config(self):
        self._var_data = defaultdict(list)
        self.substitute_library()
        self.collect_vars()
        self.ensure_experiments()
        # self.label_repeats()

    # Have to wait to writes experiment-specific XML configurations to each exp directory
    # in case a purge is happening.
    def setup_exp_configs(self):
        self.write_exp_configs()

    def substitute_library(self):
        ref_elems = self._xml_input.xpath('//*[@ref]')
        for ref_elem in ref_elems:
            ref_name = ref_elem.get('ref')
            named_elem = self._libraries.get_elem(ref_name)
            if named_elem is not None:
                merged_elem = merge_elements(named_elem, ref_elem)
                ref_elem.getparent().replace(ref_elem, merged_elem)
            else:
                logging.warning(f"No such element in library {ref_name}")

    def collect_vars(self):
        self.collect_librec_vars()
        self.collect_rerank_vars()
        self._var_coll.compute_var_configurations()

    def collect_librec_vars(self):
        value_elems = self._xml_input.xpath('/librec-auto/*[not(self::rerank)]/*/value')
        parents = [elem.getparent() for elem in value_elems]
        parents = list(set(parents))
        for parent in parents:
            vals = [elem.text for elem in parent.iterchildren(tag='value')]
            parent_path = build_parent_path(parent)
            self._var_coll.add_var('librec', parent_path, vals)

    def collect_rerank_vars(self):
        value_elems = self._xml_input.xpath('/librec-auto/rerank/*//value')
        parents = [elem.getparent() for elem in value_elems]
        parents = list(set(parents))
        for parent in parents:
            vals = [elem.text for elem in parent.iterchildren(tag='value')]
            parent_path = build_parent_path(parent)
            self._var_coll.add_var('rerank', parent_path, vals)

    # Write versions of the config file in which the parameters with multiple values are replaced with
    # a single value
    def write_exp_configs(self):
        configs = list(zip(self.get_files().get_sub_paths_iterator(),
                           iter(self._var_coll.var_confs)))
        i = 0
        for exp, vconf in configs:
            vconf.exp_no = i
            vconf.exp_dir = exp.subexp_name
            self.write_exp_config(exp, vconf)

    def write_exp_config(self, exp, vconf):
        new_xml = copy.deepcopy(self._xml_input)
        # Remove libraries. All substitutions have already happened.
        for lib in new_xml.xpath('/librec-auto/library'):
            lib.getparent().remove(lib)

        pat = re.compile(ConfigCmd._PARAM_NAME_PATH_RE)
        for vinfo in vconf.vars:
            var_elem = new_xml.xpath(vinfo.path)[0]
            var_elem.clear()
            var_elem.text = str(vinfo.val)
            var_elem.set("var", "true")
            if var_elem.tag == 'param':                     # params are distinguished by name attributes
                mat = pat.match(vinfo.path)
                var_elem.attrib['name'] = mat.group(1)
        new_xml.append(etree.Comment('This configuration file was automatically generated by librec-auto. ' +
                        'Editing may produce unpredictable results and is not recommended.'))
        outpath = exp.get_path('conf') / Files.DEFAULT_CONFIG_FILENAME
        logging.info('Writing config file ' + str(outpath))
        new_xml.getroottree().write(outpath.absolute().as_posix(), pretty_print=True)

        if vconf.ref_config is None:
            props = LibrecProperties(new_xml, self._files)
            exp.add_to_config(props.properties, 'result')
            props.save(exp)
        else:
            path = exp.get_ref_exp_flag_path()
            with path.open(mode='w') as fh:
                fh.write(vconf.ref_config.exp_dir)
                fh.write('\n')

    def has_rerank(self):
        rerank_elems = self._xml_input.xpath('/librec-auto/rerank')
        return len(rerank_elems) > 0

    def has_post(self):
        post_elems = self._xml_input.xpath('/librec-auto/post')
        return len(post_elems) > 0

    def is_valid(self):
        return self._xml_input is not None

    def thread_count(self):
        thread_elems = self._xml_input.xpath('/librec-auto/thread-count')
        if len(thread_elems) == 0:
            return 1
        else:
            return int(thread_elems[0].text)

    def setup_libraries(self):
        if utils.safe_xml_path(self._xml_input, ['librec-auto', 'library']):
            lib_elems = utils.force_list(utils.extract_from_path(self._xml_input, ['librec-auto', 'library']))
            for lib in lib_elems:
                lib_path = self.extract_library_path(lib)
                lib = ConfigLib(lib_path)
                self._libraries.add_lib(lib)

    def extract_library_path(self, lib_elem):
        file_name = lib_elem['#text'] if type(lib_elem) is OrderedDict else lib_elem
        file_path = Path(file_name)
        path_prefix = None
        if type(lib_elem) is OrderedDict:
            if '@src' in lib_elem:
                if 'system' == lib_elem['@src']:
                    path_prefix = self._files.get_lib_path()
                else:
                    print(f'librec-auto: WARNING Path source {lib_elem["@src"]} is unknown. Possible values are: system')
        else: # If library path is just as string without directory information, assume conf directory
            if file_path.parent == Path('.'):
                path_prefix = self._files.get_config_path().parent
        if path_prefix is None:
            return file_path
        else:
            return path_prefix / file_path

def read_config_file(config_file, target):
    config = ConfigCmd(config_file, target)
    if config.is_valid():
        config.load_libraries()
        config.process_config()
    return config




import datetime
from pathlib import Path
from librec_auto.core.util import xml_load_from_path, extract_from_path, force_list, SubPaths, LogFile

# A .status file looks like this
# <librec-auto-status>
#    <message>Completed</message>
#    <exp-no>1</exp-no>\
#    <param><name>rec.neighbors.knn.number</name><value>30</value></param>
#    <date>June 28, 11:00 PM</date>
# </librec-auto-status>
# TODO: Rewrite with lxml. This is kind of embarrassing.

class Status():

    _HEADER = '<?xml version="1.0"?>\n<!-- DO NOT EDIT. File automatically generated by librec-auto -->\n'
    _TEMPLATE_FRONT = '<librec-auto-status>\n<message>{}</message>\n<exp-no>{}</exp-no>\n<date>{}</date>\n'
    _TEMPLATE_END = '</librec-auto-status>\n'
    _TEMPLATE_LINE = '<param><name>{}</name><value>{}</value></param>\n'

    def __init__(self, sub_paths):
        self._subpaths = sub_paths
        status_path = self._subpaths.get_path('status')

        if status_path.exists():
            self._name = sub_paths.subexp_name
            self._status_xml = xml_load_from_path(status_path)
            self._message = extract_from_path(self._status_xml, ['librec-auto-status', 'message'])
            if self._subpaths.get_path('log').exists():
                self._log = LogFile(self._subpaths)
            else:
                self._log = None
            params = extract_from_path(self._status_xml, ['librec-auto-status', 'param'])
            if params != None:
                self.process_params(force_list(params))
            else:
                self._params = []
                self.m_vals = []

    def __str__(self):
        params_string = self.get_params_string()
        if self._log:
            results_string = self.get_log_info()
        else:
            results_string = "No LibRec results"
        return f'Status({self._name}:{self._message}{params_string} Overall{results_string})'

    def is_completed(self):
        return self._message == 'Completed'

    def process_params(self, param_xml):
        param_list = []
        val_list = []

        for pdict in param_xml:
            param_list.append(pdict['name'])
            val_list.append(pdict['value'])

        self._params = param_list
        self._vals = val_list

    def get_params_string(self):
        params_string = ''
        if self._params == []:
            return " No parameters"
        for param, val in zip(self._params, self.m_vals):
            params_string += f' {param}: {val}'
        return params_string

    def get_log_info(self):
        kcv_count = self._log.get_kcv_count()
        if kcv_count is None:
            return self.get_metric_info(self._log, 0)
        else:
            return self.get_metric_info(self._log, -1)

    def get_metric_info(self, log, idx):
        metric_info = ''
        for metric_name in log.get_metrics():
            metric_value = log.get_metric_values(metric_name)[idx]
            metric_info = metric_info + f' {metric_name}: {float(metric_value):.3f}'
        return metric_info


    # Accept list of vars and tuples
    @staticmethod
    def save_status(msg, exp_count, config, paths):
        status_file = paths.get_path('status')
        status_front = Status._TEMPLATE_FRONT.format(msg, exp_count, datetime.datetime.now())

        status_params = ''
        conf_xml = config.get_files().get_sub_paths(exp_count).get_exp_conf()
        var_elems = conf_xml.xpath("//*[@var='true']")
        for var_elem in var_elems:
            if var_elem.tag == 'param':
                var_name = var_elem.get('name')
            else:
                var_name = var_elem.tag
            var_value = var_elem.text

            status_params = status_params + Status._TEMPLATE_LINE.format(var_name, var_value)

        status_info = Status._HEADER + status_front + status_params + Status._TEMPLATE_END

        with status_file.open(mode='w') as fh:
            fh.write(str(status_info))



from librec_auto.core.cmd import Cmd
from librec_auto.core.util import Files, SubPaths, Status
from librec_auto.core import ConfigCmd
import os
import subprocess
import shlex
import  time
from pathlib import Path, WindowsPath

class LibrecCmd (Cmd):

    _DEFAULT_WRAPPER_CLASS = "net.that_recsys_lab.auto.SingleJobRunner"

    def __str__(self):
        return f'LibrecCmd(sub-exp: {self._sub_no}, command: {self._command})'

    def __init__(self, command, sub_no):
        self._command = command
        self._sub_no = sub_no
        self._config: ConfigCmd = None
        self._sub_path: SubPaths = None

    def setup(self, args):
        pass

    # 2020-01-06 RB Theoretically, subprocess.run is the right way to do this, but capturing the log output
    # seems to work more naturally with Popen. A mystery for future developers. Also, capture_output requires
    # Python 3.8, which may be too much to ask at this point.
    # proc = subprocess.run(cmd, capture_output=True
    def execute_librec(self):
        cmd = self.create_proc_spec()

        if len(cmd) == 0:
            print ("librec-auto: Unknown command {self._command}. Skipping LibRec execution.")
            self.status = Cmd.STATUS_ERROR
            return

        print(f"librec-auto: Running librec. {cmd}")
        log_path = self._sub_path.get_log_path()

        f = open(str(log_path), 'w+')
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        for line in p.stdout:
            line_string = str(line, 'utf-8')
            f.write(line_string)
            print(line_string, end='')
        f.close()

        #p.wait()

        if type(p.returncode) is 'int' and p.returncode < 0:
            self.status = Cmd.STATUS_ERROR
        else:
            self.status = Cmd.STATUS_COMPLETE


    def dry_run_librec(self):
        cmd = self.create_proc_spec()

        proc_spec = ' '.join(cmd)
        print (f'librec-auto (DR): Executing librec command: {self},  sub-exp: {self._sub_path.subexp_name}, exec: {proc_spec}')
        # Only for testing parallel function
        # time.sleep(1.0)

    def dry_run(self, config):
        self._config = config
        self._sub_path = config.get_files().get_sub_paths(self._sub_no)
        link = self._sub_path.get_ref_exp_name()
        if not link:
            self.dry_run_librec()
        else:
            print(f'librec-auto (DR): Skipping librec. Getting results from {link}')

    def execute(self, config: ConfigCmd):
        self._config = config
        self._sub_path = config.get_files().get_sub_paths(self._sub_no)
        if not self._sub_path.get_ref_exp_name():
            self.ensure_clean_log()

            Status.save_status("Executing", self._sub_no, config, self._sub_path)
            self.execute_librec()
        else:
            self.status = Cmd.STATUS_COMPLETE
        Status.save_status("Completed", self._sub_no, config, self._sub_path)

    # Checks for any contents of split directory, which would have been removed by purging
    def split_exists(self):
        split_path = self._config.get_files().get_split_path()
        return split_path.exists() and any(os.scandir(split_path))

    # Checks for any contents of results directory, which would have been removed by purging
    def results_exist(self):
        sub_paths = self._config.get_files().get_sub_paths(self._sub_no)
        results_path = sub_paths.get_path('results')
        return any(os.scandir(results_path))

    # log file appends by default
    def ensure_clean_log(self):
        librec_log = log_path = self._sub_path.get_path('log') / SubPaths.DEFAULT_LOG_FILENAME
        if librec_log.is_file():
            librec_log.unlink()

    def create_proc_spec(self):
        classpath = self._config.get_files().get_classpath()
        mainClass = self._DEFAULT_WRAPPER_CLASS
        confpath = self._sub_path.get_librec_properties_path()
        confpath_str = str(confpath)

        java_command = self.select_librec_action()
        if java_command is None:
            return []
        else:
            return ['java', '-cp', classpath, mainClass, confpath_str, java_command]

    # 2019-11-23 RB Not sure if this step can be replaced by more checking when commands are created.
    def select_librec_action(self):
        expVal = self._sub_path.subexp_name

        if self._command == 'split':
            # check if split exists, if so split command doesn't make sense. Does not purge here.
            if self.split_exists():
                print("Split already exists. Skipping.")
                return None
            else:
                return 'split'

        if self._command == 'eval':
            if self.results_exist():
                return 'reRunEval'
            else:  # No result file present, Then check if split exists
                if self.split_exists():
                    return 'exp-eval'
                else:
                    return 'full'

        if self._command == 'full':
            return 'full'

        if self._command == 'check':
            return 'check'
        else:
            return None



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="librec-auto",
    version="0.1.28",
	scripts=['librec_auto\__main__.py'] ,
    author="Masoud Mansoury and Robin Burke",
    author_email="masoodmansoury@gmail.com",
    description="The librec-auto project aims to automate recommender system experimens using LibRec.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/that-recsys-lab/librec-auto",
    packages=setuptools.find_packages(),
	include_package_data=True,
	install_requires=[
          'matplotlib',
		  'pandas',
		  'numpy',		  
    ],
	classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from librec_auto.core.util import SubPaths
import re
import os

# By default, the most recent one in the directory
class LogFile:

    def __str__(self):
        return f'LogFile({self._values}'

    def __init__(self, paths: SubPaths):
        self._metrics = []
        self._values = {}

        self._log_path = self.newest_log(paths)


        self.parse_log()

    def newest_log(self, paths):
        log_dir = paths.get_path('log')
        log_files = os.listdir(log_dir)
        newest_file = sorted(log_files, reverse=True)[0]
        return log_dir / newest_file

    def get_metric_values(self, metric):
        if metric in self._values:
            return self._values[metric]
        else:
            return []

    def get_metrics(self):
        return self._metrics

    def get_metric_count(self):
        return len(self._metrics)

    def add_metric_value(self, metric, value):
        if metric in self._metrics:
            self._values[metric].append(value)
        else:
            self._values[metric] = [value]
            self._metrics.append(metric)

    def get_kcv_count(self):
        return self._kcv

    def parse_log(self):
        eval_pattern_str = r'.*Evaluator info:(.*)Evaluator is (-?\d+.?\d+)'
        kcv_pattern_str = r'.*Splitting .* on fold (\d+)'
#        kcv_pattern_str = r'.*Splitter info: .* times is (\d+)'
        final_pattern_str = r'.*Evaluator value:(.*)Evaluator is (-?\d+.?\d+)'

        eval_pattern = re.compile(eval_pattern_str)
        kcv_pattern = re.compile(kcv_pattern_str)
        final_pattern = re.compile(final_pattern_str)

        with open(str(self._log_path), 'r', newline='\n') as fl:

            i = 0
            for ln in fl:
                i += 1
                eval = re.match(eval_pattern, ln)
                kcv = re.match(kcv_pattern, ln)
                final = re.match(final_pattern, ln)

                # A little bit of a hack. The average summary value is added at the end of the list
                if eval is not None or final is not None:
                    if eval is None:
                        eval = final
                    metric_name = eval.group(1)
                    metric_value = eval.group(2)
                    self.add_metric_value(metric_name, metric_value)

            if kcv is not None:
                self._kcv = kcv.group(1)

